from K2600acceslib.kth2600 import K2600
from . import logger

logger = logger.getChild(__name__)


class K2600Probecard(K2600):
    """ Developing test for Probe card using keithley 2636B """

    def __init__(self, source_str, verbose=False):
        super().__init__(source_str)
        self.verbose = verbose

    def main(self, a_l, b_l, rnge, level, avdd_v_c, avdd_c_c, vdd_v_c, vdd_c):
        """
        Execute probecard main test.
        :param a_l: chan A source limit current.
        :param b_l: chan B source limit current.
        :param rnge: chan A and B source range, can be 20, 200, .... check K2600 manual.
        :param level: chan A and B source level, can be 20, 200, .... check K2600 manual.
        :param avdd_v_c: AVDD voltage test comparison threshold.
        :param avdd_c_c: AVDD current test comparison threshold.
        :param vdd_v_c: VDD voltage test comparison threshold.
        :param vdd_c: VDD current test comparison threshold.
        :return: Test output binary array. shape (4,).
        """
        self.clear_reset()
        self.set_values(a_l, b_l, rnge, level)
        self.set_k_out_on()
        test_output_bin, test_output = self.read_compare(avdd_v_c, avdd_c_c, vdd_v_c, vdd_c)
        return test_output_bin, test_output

    def clear_reset(self):
        # Check and clear error
        self.clear_connection()
        self.reset_source_meter()
        self.clearErrorQueue()

    def set_values(self, a_l, b_l, rnge, level):
        """ Set all values """
        self.scpi_src_out_dcv("a")
        self.scpi_src_out_dcv("b")
        self.scpi_src_limit("a", "i", a_l)
        self.scpi_src_limit("b", "i", b_l)
        self.scpi_src_range("a", "v", rnge)
        self.scpi_src_range("b", "v", rnge)
        self.scpi_src_level("a", "v", level)
        self.scpi_src_level("b", "v", level)
        self.checkIfError()

        # Read and compare values
        assert self.scpi_src_read_limit("a", "i") == a_l
        assert self.scpi_src_read_limit("b", "i") == b_l
        assert self.scpi_src_read_range("a", "v") == rnge
        assert self.scpi_src_read_range("b", "v") == rnge
        assert self.scpi_src_read_level("a", "v") == level
        assert self.scpi_src_read_level("b", "v") == level
        self.checkIfError()

    def read_compare(self, avdd_v_c, avdd_c_c, vdd_v_c, vdd_c):
        """ Read output values and compare """
        # Readout
        avdd_voltage = self.scpi_measure_read("a", "v")
        avdd_current = self.scpi_measure_read("a", "i")
        vdd_voltage = self.scpi_measure_read("b", "v")
        vdd_current = self.scpi_measure_read("b", "i")
        self.checkIfError()

        # Compare
        iv0 = avdd_voltage >= avdd_v_c
        iv1 = avdd_current >= avdd_c_c
        iv2 = vdd_voltage >= vdd_v_c
        iv3 = vdd_current >= vdd_c

        if self.verbose:
            logger.info(f"Readout values: {avdd_voltage}, {avdd_current}, {vdd_voltage}, {vdd_current}")
            logger.info(f"Readout bin result: {int(iv0)}{int(iv1)}{int(iv2)}{int(iv3)}")

        if iv0 == 0 or iv2 == 0:
            self.set_k_out_off()

        return [int(iv0), int(iv1), int(iv2), int(iv3)], [avdd_voltage, avdd_current, vdd_voltage, vdd_current]

    def set_k_out_on(self):
        """ Setting both channels of keithley off"""
        # Set both output on.
        self.scpi_src_output("a", "ON")
        self.scpi_src_output("b", "ON")
        self.checkIfError()

    def set_k_out_off(self):
        """ Setting both channels of keithley off"""
        self.scpi_src_output("a", "OFF")
        self.scpi_src_output("b", "OFF")
        self.checkIfError()



import logging.handlers

# Defined here since it is imported in other pyvisa modules
BASE_NAME = 'k2600'
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

formatter = logging.Formatter(
    "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)


def log_to_screen(level=logging.DEBUG) -> None:
    log_to_stream(None, level)  # sys.stderr by default


def log_to_stream(stream_output, level=logging.DEBUG) -> None:
    logger.setLevel(level)
    ch = logging.StreamHandler(stream_output)
    ch.setLevel(level)
    ch.setFormatter(formatter)

    logger.addHandler(ch)


def log_add_file_handler(file_path, level=logging.DEBUG):
    fh = logging.handlers.RotatingFileHandler(filename=file_path, maxBytes=5 * 1024 * 1024, backupCount=5)
    fh.setLevel(level=level)
    fh.setFormatter(formatter)
    logger.addHandler(fh)


__version__ = "unknown"

" It is a list of strings defining what symbols in a module willbe exported when"
" from <module> import * is used on the module."

__all__ = [
    "logger",
    "log_to_screen",
    "log_to_stream",
    "log_add_file_handler",
]



import pyvisa as visa
from . import logger

logger = logger.getChild(__name__)


class K2600:
    """
    Control for KEITHLEY 2410 using GPIB.
    """

    def __init__(self, source_str):
        self._inst = None
        self._source_str = source_str
        self._init_connection()
        self.clear_connection()

    def _init_connection(self):
        """
        Initialize connection with KEITHLEY and check if it is possible to identify the hardware.
        """
        try:
            rm = visa.ResourceManager()
            self._inst = rm.open_resource(self._source_str)
            logger.info(self._inst.query("*IDN?"))

        except visa.errors.VisaIOError:
            logger.error("Error Initializing connection, check input string")
            raise ConnectionError

    def close_connection(self):
        """
        Closes the VISA session and marks the handle as invalid.
        """
        self._inst.close()

    def clear_connection(self):
        """
        Clear this resource.
        """
        self._inst.clear()

    def reset_source_meter(self):
        """ Returns the SourceMeter to default conditions. """
        self._inst.write("*RST")

    def scpi_reset(self, chan):
        """
        Reset values in default values
        """
        assert (chan == "a" or chan == "b")
        self._inst.write(f'smu{chan}.reset()')

    def scpi_src_out_dcv(self, chan):
        """
        Set the source output to OUTPUT_DCVOLTS
        """
        assert (chan == "a" or chan == "b")
        self._inst.write(f'smu{chan}.source.func = smu{chan}.OUTPUT_DCVOLTS')

    def scpi_src_out_dca(self, chan):
        """
        Set the source output to OUTPUT_DCAMPS
        """
        assert (chan == "a" or chan == "b")
        self._inst.write(f'smu{chan}.source.func = smu{chan}.OUTPUT_DCAMPS')

    def scpi_src_range(self, chan, s_type, value):
        """
        Set the source range value
        """
        assert (chan == "a" or chan == "b")
        assert (s_type == "v" or s_type == "i")
        self._inst.write(f'smu{chan}.source.range{s_type} = {value}')

    def scpi_src_level(self, chan, s_type, value):
        """
        Set the source level value
        """
        assert (chan == "a" or chan == "b")
        assert (s_type == "v" or s_type == "i")
        self._inst.write(f'smu{chan}.source.level{s_type} = {value}')

    def scpi_measure_range(self, chan, s_type, value):
        """
        Set the measure range value
        """
        assert (chan == "a" or chan == "b")
        assert (s_type == "v" or s_type == "i")
        self._inst.write(f'smu{chan}.measure.range{s_type} = {value}')

    def scpi_src_limit(self, chan, s_type, value):
        """
        Set the source limit value
        """
        assert (chan == "a" or chan == "b")
        assert (s_type == "v" or s_type == "i")
        self._inst.write(f'smu{chan}.source.limit{s_type} = {value}')

    def scpi_src_output(self, chan, state):
        """
        Set the source limit value
        """
        assert (chan == "a" or chan == "b")
        assert (state == "ON" or state == "OFF")
        self._inst.write(f'smu{chan}.source.output = smu{chan}.OUTPUT_{state}')

    # ***************************** Read ***************************** #
    def scpi_src_read_limit(self, chan, s_type):
        """
        Read the source limit value
        """
        assert (chan == "a" or chan == "b")
        assert (s_type == "v" or s_type == "i")
        self._inst.write(f'limit = smu{chan}.source.limit{s_type}')
        return float(self._inst.query("print(limit)"))

    def scpi_src_read_range(self, chan, s_type):
        """
        Read the source limit value
        """
        assert (chan == "a" or chan == "b")
        assert (s_type == "v" or s_type == "i")
        self._inst.write(f'range = smu{chan}.measure.range{s_type}')
        return float(self._inst.query("print(range)"))

    def scpi_src_read_level(self, chan, s_type):
        """
        Read the source limit value
        """
        assert (chan == "a" or chan == "b")
        assert (s_type == "v" or s_type == "i")
        self._inst.write(f'level = smu{chan}.source.level{s_type}')
        return float(self._inst.query("print(level)"))

    def scpi_measure_read(self, chan, s_type):
        """
        Read the source limit value
        """
        assert (chan == "a" or chan == "b")
        assert (s_type == "v" or s_type == "i")
        try:
            self._inst.write(f'reading = smu{chan}.measure.{s_type}()')
            return float(self._inst.query("print(reading)"))
        except visa.errors.VisaIOError:
            logger.error("Impossible to read, time out occurred in measure reading")
            return -1

    # ***************************** Error ***************************** #
    def clearErrorQueue(self):
        self._inst.write("*CLS")

    def checkIfError(self):
        try:
            try:
                err = int(self._inst.query("*ESR?"))
            except visa.errors.VisaIOError:
                logger.error("pyvisa.errors.VisaIOError: Timeout expired before operation completed.")
                return -1

            err &= ~(1 << 7)  # Delete Power On bit
            err &= ~(1 << 0)  # Delete OPC bit
            if err != 0:
                if err & (1 << 6):
                    logger.debug("User request error detected")
                if err & (1 << 5):
                    logger.debug("Command error detected")
                if err & (1 << 4):
                    logger.debug("Execution error detected")
                if err & (1 << 3):
                    logger.debug("Device-Dependent error detected")
                if err & (1 << 2):
                    logger.debug("Query error detected")
            else:
                logger.info("No error in keithley operation.")
        except ValueError:
            logger.error("The output queue is full and it return data not releted with '*ESR?' command.")
            return -1

        return err
